let drawModeMarker = "marker";
let DrawModeCircle = "circle";
let DrawModeArea = "area";
let DrawModeFitBounds = "fitbounds";
let DrawModeReset = "reset";

function DrawingManager(map, opts) {

    try {
        google;
    } catch (e) {
        throw Error('Google Map JS API is not ready yet!');
    }

    let me = this;
    me.extend(DrawingManager, google.maps.OverlayView);

    me.map = map;
    me._opts = opts;
    me._drawingType = null;
    me.markerOptions = opts.markerOptions || {};
    me.circleOptions = opts.circleOptions || {};
    me.areaOptions = opts.areaOptions || {};
    me.isDraw = false;
    me.radius = opts.circleOptions.radius;
    me.vertex = [];
    me._circle = null;
    me._centerMarker = null;
    me._area = null;
    me.vertexCount = 0;
    me.polygonArea = new Map();
    me.isMoreArea = opts.areaOptions.isMoreArea || false;
    me.squareArea = opts.areaOptions.areaMaxSquare || 1400;
    me.isRedraw = null;
    me.mode = opts.activeMode;

};

DrawingManager.prototype.onAdd = function () {
    const me = this;
    google.maps.event.addListener(me.map, 'click', function (event) {
        google.maps.event.trigger(me, 'click', event);
    });

};

DrawingManager.prototype.setDrawingMode = function (drawingType, opts) {
    if (!google || !google.maps) { return; }
    const me = this;
    me.markerOptions = opts.markerOptions || {};
    me.circleOptions = opts.circleOptions || {};
    me.areaOptions = opts.areaOptions || {};
    me.isDraw = opts.isDraw;
    me.isRedraw = opts.isRedraw;
    me.radius = opts.circleOptions.radius;
    me.mode = opts.activeMode;

    me.markerOptions && me.setPosition(me.markerOptions.markerLatitude, me.markerOptions.markerLongitude);

    const info = {
        isRedraw: opts.isRedraw,
        isDraw: opts.isDraw,
        mode: me.mode,
    }

    google.maps.event.trigger(me, 'draw:mode', info);


    if (me.areaOptions) {
        me.areaCoordinates = me.areaOptions.areaCoordinates;
    }

    switch (drawingType) {
        case drawModeMarker:
            me._bindMarker(false);
            break;
        case DrawModeCircle:
            me._bindMarker(true);
            break;
        case DrawModeArea:
            me._bindArea();
            break;
        case DrawModeFitBounds:
            me._fitBounds();
            break;
        case DrawModeReset:
            me.reset();
            break;
        default:
            me._remove();
            break;
    }
    me._drawingType = drawingType;

},

    DrawingManager.prototype.reset = function () {
        const me = this;
        google.maps.event.clearListeners(me, 'click');
        google.maps.event.clearListeners(me.map.getDiv(), 'mousedown');
        me.map.setOptions({ clickableIcons: false });
        me.isRedraw = false;
    },

    DrawingManager.prototype.remove = function () {
        const me = this;
        me.map.setOptions({ clickableIcons: false });
        me._remove();
        me.isRedraw = false;
    },

    DrawingManager.prototype.resetMarker = function () {
        const me = this;
        if (me._centerMarker) { me._removeCenterMarker(); }
    },

    DrawingManager.prototype.isAnimateMarker = function (enabled) {
        const me = this;
        const shape = {
            coords: [26, 26, enabled && 80 || 100],
            type: 'circle',
        };
        me._centerMarker && me._centerMarker.setOptions({ shape });
    },

    DrawingManager.prototype.addMarker = function (latitude, longitude) {
        const me = this;
        me.setPosition(latitude, longitude);
        me.isDraw = false;
        me._bindMarker(false);
    },

    DrawingManager.prototype._fitBounds = function () {
        const me = this;
        if (me._centerMarker) {
            me._fitBoundsMarker();
        }
        if (me._area) {
            me._fitBoundsArea();
        }
        if (me._circle) {
            me._fitBoundsCircle();
        }
    },

    DrawingManager.prototype._fitBoundsMarker = function () {
        let me = this;
        if (!me._centerMarker) { return; }
        me.map.getZoom() > 11 ?
            me.map.setZoom(me.map.getZoom())
            : me.map.setZoom(11);
        me.map.setCenter(me._centerMarker.getPosition());
    }

DrawingManager.prototype._fitBoundsCircle = function () {
    const me = this;
    if (!me._circle) { return; }
    me.map.fitBounds(me._circle.getBounds());
    google.maps.event.trigger(me, 'draw:circle_create', this._getInfo());
},

    DrawingManager.prototype._setPosition = function (e) {
        const me = this;
        me.position = null;

        if (e) {
            me.position = e.latLng;
        }
    },

    DrawingManager.prototype.setPosition = function (lat, lng) {
        let me = this;
        me.position = null;
        me.position = new google.maps.LatLng(lat, lng);

    },

    DrawingManager.prototype._remove = function () {
        const me = this;
        if (me._circle) { me._removeCircle(); }
        if (me._centerMarker) { me._removeCenterMarker(); }
        if (me._area || me.polygonArea.length) { me._removeArea(); }
        google.maps.event.clearListeners(me, 'click');
    },

    DrawingManager.prototype._bindArea = function () {
        let me = this;
        google.maps.event.clearListeners(me, 'click');
        me.map.setOptions({ clickableIcons: false });
        let pathcLat = [];
        me._removeCircle();
        me._removeCenterMarker();

        !me.isMoreArea && me._removeArea();

        if (me.isMoreArea && me.polygonArea.size === 5) { return false };

        var createArea = () => {
            me._setDrawing(false);
            const polygonOptions = {
                map: me.map,
                strokeColor: me.areaOptions.strokeColor,
                strokeOpacity: me.areaOptions.strokeOpacity,
                fillColor: me.areaOptions.fillColor,
                fillOpacity: me.areaOptions.fillOpacity,
                strokeWeight: me.areaOptions.strokeWeight,
                clickable: false,
                editable: false
            }

            me._area = new google.maps.Polyline(polygonOptions);

            var move = google.maps.event.addListener(me.map, 'mousemove', function (e) {
                pathcLat.push(e.latLng);
                if (e.latLng) {
                    let area = google.maps.geometry.spherical.computeArea(pathcLat);
                    me.map.setOptions({ draggableCursor: 'not-allowed' });
                    if (area < me.squareArea * 1000000) {
                        me._area.getPath().push(e.latLng);
                        me.map.setOptions({ draggableCursor: 'crosshair' });
                        return false;
                    }
                }
            });

            var moveUp = google.maps.event.addListenerOnce(me.map, 'mouseup', () => {
                google.maps.event.removeListener(move);
                google.maps.event.removeListener(moveUp);
                me.map.setOptions({ draggableCursor: 'crosshair' });
                var path = me._area.getPath();
                me._area.setMap(null);
                me._setDrawing(true);
                if (!me._area.getPath().length) { return false }

                const opts = {
                    map: me.map,
                    strokeColor: me.areaOptions.strokeColor,
                    strokeOpacity: me.areaOptions.strokeOpacity,
                    fillColor: me.areaOptions.fillColor,
                    fillOpacity: me.areaOptions.fillOpacity,
                    strokeWeight: me.areaOptions.strokeWeight,
                    clickable: false,
                    path: path,
                    editable: false
                }

                me.vertexCount++;
                me.polygonArea.set(me.vertexCount, me._area);
                me._area = new google.maps.Polygon(opts);
                google.maps.event.clearListeners(me.map.getDiv(), 'mousedown');
                me.map.setOptions({ draggableCursor: me.map.get('draggableCursor') });

                me._fitBoundsArea();

                if (me.isMoreArea) {
                    patchAreaLatLng = me._area.getPath().getArray();
                    let patchAreaLatLngMax = me.findRoundPoints(patchAreaLatLng, 0.001);
                    if (patchAreaLatLngMax) {
                        if (patchAreaLatLngMax.length == 0) {
                            patchAreaLatLngMax = me.findRoundPoints(patchAreaLatLng, 0.01);
                        }
                        if (patchAreaLatLngMax.length == 0) {
                            patchAreaLatLngMax = me.findRoundPoints(patchAreaLatLng, 1);
                        }
                        if (patchAreaLatLngMax.length == 0) {
                            patchAreaLatLngMax = patchAreaLatLng;
                        }
                    }

                    var pointSelect = null;
                    var pointSelect = patchAreaLatLngMax[0];

                    for (var n = 0; n < patchAreaLatLngMax.length; n++) {
                        let item = patchAreaLatLngMax[n];
                        if (item.lng() > pointSelect.lng()) {
                            pointSelect = item;
                        }
                    }

                    me.createVertexArea(pointSelect);
                }
            });
        }

        if (this.isDraw) {
            google.maps.event.addDomListener(me.map.getDiv(), 'mousedown', (e) => {
                e = window.event || e;
                if (e.button == 2 || e.detail > 2) { return; }
                createArea();
            });
        }
        else {
            me._redrawArea();
        }
    },

    DrawingManager.prototype.createVertexArea = function (pointSelect) {
        me = this;
        _vertex = new google.maps.Marker({
            position: pointSelect,
            map: me.map,
            zIndex: 1000002,
            draggable: false,
            icon: { url: me.areaOptions.areaIconVertex, scaledSize: new google.maps.Size(24, 24) },
            raiseOnDrag: false,
            optimized: true,
            id: me.vertexCount,

        });

        _vertex.addListener('click', function () {
            this.setMap(null);
            if (me.polygonArea.get(this.id)) {
                me.polygonArea.get(this.id).getPath().clear();
                me.polygonArea.get(this.id).setMap(null);
                google.maps.event.trigger(me, 'draw:area_remove', me._convertCoordinates());
            }
        });

        me.vertex.push(_vertex);

    }

DrawingManager.prototype.findRoundPoints = function (patch, round) {

    var bounds = new google.maps.LatLngBounds();
    patch.forEach(function (element, index) { bounds.extend(element) });
    var center = bounds.getCenter();
    var result = [];
    var roundMinLat = center.lat() - round;
    var roundMaxLat = center.lat() + round;
    for (var n = 0; n < patch.length; n++) {
        let item = patch[n];
        if (item.lat() >= roundMinLat && item.lat() <= roundMaxLat) {
            if (item.lng() > center.lng()) {
                result.push(item);
            }
        }
    }

    return result;
},

    DrawingManager.prototype._redrawArea = function () {
        const me = this;
        google.maps.event.clearListeners(me, 'mousedown');
        if (me.areaCoordinates) {
            const latlngs = me._convertlatlngs();
            const opts = {
                map: me.map,
                strokeColor: me.areaOptions.strokeColor,
                strokeOpacity: me.areaOptions.strokeOpacity,
                fillColor: me.areaOptions.fillColor,
                fillOpacity: me.areaOptions.fillOpacity,
                strokeWeight: me.areaOptions.strokeWeight,
                clickable: false,
                path: latlngs,
                editable: false
            }
            me._area = new google.maps.Polygon(opts);
            me._fitBoundsArea();
        }
    },

    DrawingManager.prototype._convertlatlngs = function () {
        const me = this;
        let latlngs = [];


        if (me.areaCoordinates.length > 0) {
            for (var n = 0; n < me.areaCoordinates.length; n++) {
                let item = me.areaCoordinates[n];
                latlngs.push(new google.maps.LatLng(item.latitude, item.longitude));
            }
            return latlngs;
        }
    },

    DrawingManager.prototype._setDrawing = function (enabled) {
        const me = this;
        me.map.setOptions({
            draggable: enabled,
            scrollwheel: enabled,
            disableDoubleClickZoom: enabled,
        });
    },

    DrawingManager.prototype._fitBoundsArea = function () {
        const me = this;
        if (me._area) {
            let patchLatLng = me._area.getPath().getArray()
            let bounds = new google.maps.LatLngBounds();
            for (var n = 0; n < patchLatLng.length; n++) {
                bounds.extend(patchLatLng[n]);
            }
            me.map.panToBounds(bounds);
            me.map.fitBounds(bounds);
        }
        google.maps.event.trigger(me, 'draw:area_create', me._convertCoordinates(me.vertexCount));
    },

    DrawingManager.prototype._convertCoordinates = function (cur) {
        let me = this;
        let bounds = new google.maps.LatLngBounds();
        me._area.getPath().getArray().forEach(function (element, index) { bounds.extend(element) });
        let _center = bounds.getCenter();
        let center = {};
        let positions = [];
        let position = [];
        let arrayArea = {};
        let id = null;
        if (!me.isMoreArea) {
            let latLng = [];
            for (var n = 0; n < me._area.getPath().getArray().length; n++) {
                let item = me._area.getPath().getArray()[n];
                let position = {
                    latitude: item.lat(),
                    longitude: item.lng(),
                }
                latLng.push(position);
                position = positions;
                id = 0;
            }
            positions.push(latLng);
            position = positions[0];
            center = {
                latitude: _center.lat(),
                longitude: _center.lng(),
            }

            arrayArea = {
                id: id,
                positions,
                position,
                center,
                zoom: me.map.getZoom(),
            }
            return arrayArea;
        }
        else {
            me.polygonArea.forEach((key, value) => {
                let latLng = [];
                for (var n = 0; n < key.getPath().getArray().length; n++) {
                    let item = key.getPath().getArray()[n];
                    let position = {
                        latitude: item.lat(),
                        longitude: item.lng(),
                    }
                    latLng.push(position);
                    id = value;
                }
                positions.push(latLng);
                if (cur) {
                    if (id === cur) {
                        position = positions[cur - 1];
                    }
                }
            });

            arrayArea = {
                id: id,
                positions,
                position
            }
        }

        return arrayArea;
    },


    DrawingManager.prototype._bindMarker = function (enabledCircle) {
        const me = this;
        this.isDraw && !enabledCircle && me.map.setOptions({ clickableIcons: true });
        me._removeCircle();
        me._removeArea();
        me._removeCenterMarker();
        google.maps.event.clearListeners(me, 'click');
        google.maps.event.clearListeners(me, 'mousedown');
        google.maps.event.clearListeners(me, 'mouseup');
        google.maps.event.clearListeners(me.map.getDiv(), 'mousedown');

        var createCenterMarker = (e) => {
            me._removeCircle();
            me._removeCenterMarker();

            e && me._setPosition(e);

            if (me.position) {
                me._centerMarker = new google.maps.Marker({
                    draggable: false,
                    visible: me.markerOptions.isVisibleMarker || true,
                    zIndex: 1000002,
                    position: me.position,
                    icon: enabledCircle ? me.circleOptions.iconUrl : me.markerOptions.iconUrl,
                    raiseOnDrag: false,
                    optimized: false,
                    placeId: e && e.placeId || ''
                });

                me._centerMarker.setMap(me.map);

                me.markerOptions.fitBounds && !enabledCircle && me._fitBoundsMarker();
                me.markerOptions.isMarkerAnimate && !enabledCircle && me.isAnimateMarker(true);

                if (!enabledCircle) { google.maps.event.trigger(me, 'draw:marker_create', me._getPosition()); }

                google.maps.event.addListener(me._centerMarker, 'click', () => {
                    google.maps.event.trigger(me, 'draw:marker_click', me._getPosition());
                });
                google.maps.event.addListener(me._centerMarker, 'mouseover', function (event) {
                    const pixel = me.getXYbyEvent(event);
                    const pos = me._centerMarker.getPosition();
                    const position = {
                        latitude: pos.lat(),
                        longitude: pos.lng(),
                    }
                    const ev = {
                        clientX: pixel.clientX,
                        clientY: pixel.clientY,
                        position
                    }
                    google.maps.event.trigger(me, 'draw:marker_mouseover', ev);
                });

                google.maps.event.addListener(me._centerMarker, 'mouseout', function (event) {
                    google.maps.event.trigger(me, 'draw:marker_mouseout');
                });


                me._centerMarker.setCursor('pointer');

                (enabledCircle && me._centerMarker) && me._bindCircle();
            }

        }

        if (!this.isDraw) {
            createCenterMarker();
            google.maps.event.clearListeners(me, 'click');
            google.maps.event.removeListener(me.markerListener);
        }

        me.markerListener = google.maps.event.addListener(me, 'click', (event) => {
            if (this.isDraw || me.isRedraw) {
                event.placeId && event.stop();
                createCenterMarker(event);
            }
        });
    },

    DrawingManager.prototype._bindCircle = function () {
        const me = this;

        me._removeCircle();

        if (me._centerMarker) {
            me._circle = new google.maps.Circle({
                strokeColor: me.circleOptions.strokeColor,
                strokeOpacity: me.circleOptions.strokeOpacity,
                strokeWeight: me.circleOptions.strokeWeight,
                fillColor: me.circleOptions.fillColor,
                fillOpacity: me.circleOptions.fillOpacity,
                center: me._centerMarker.getPosition(),
                radius: me.radius,
                optimized: true,
            });

            me._circle.setMap(me.map);
            me._circle.bindTo('center', me._centerMarker, 'position');

            me._fitBoundsCircle();

            me._createVertexMarker();
            me._createVertexMarkerCopy();
            me._centerMarker.setDraggable(true);
            me._centerMarker.setCursor('move');
            me._setIconMarker(me.circleOptions.iconUrl);
            me._centerMarkerAddEventListener();
        }
    },

    DrawingManager.prototype._setIconMarker = function (iconUrl) {
        const me = this;
        if (me._centerMarker) {
            const icon = {
                url: iconUrl,
            };
            me._centerMarker.setIcon(icon);
        }
    },

    DrawingManager.prototype.setEnableDraw = function (enabled) {
        this.isDraw = enabled;
    },

    DrawingManager.prototype._removeCircle = function () {
        let me = this;
        if (me._circle) {
            me._circle.setMap(null);
            me._circle = null;
            me._vertexMarker.setMap(null);
            if (me._vertexMarkerCopy) {
                me._vertexMarkerCopy.setMap(null);
            }
            google.maps.event.trigger(me, 'draw:circle_remove', null);
        }
    },

    DrawingManager.prototype._removeCenterMarker = function () {
        const me = this;
        if (me._centerMarker && !me._circle) {
            me._centerMarker.setMap(null);
            me._centerMarker = null;
        }
        google.maps.event.trigger(me, 'draw:marker_remove', null);
    },

    DrawingManager.prototype._removeArea = function () {
        const me = this;

        if (me._area) {
            me._area.setMap(null);
            me._area = null;
            google.maps.event.trigger(me, 'draw:area_remove', []);
        }

        me.polygonArea.forEach((key) => {
            key.getPath().clear();
            key.setMap(null);
            google.maps.event.trigger(me, 'draw:area_remove', []);
        });

        me.vertex.forEach((key) => {
            key.setMap(null);
        });

        me.polygonArea.clear();
        me.vertex = [];
        me.currentAreaCount = 0;

    },

    DrawingManager.prototype._createVertexMarker = function () {
        let me = this;
        me.to = null;
        me.to = me.destination(me._centerMarker.getPosition(), 90, me.radius);
        me.fillColor = 'white';

        let svg = [
            `<?xml version="1.0"?>`,
            `<svg width="16px" height="16px" viewBox="0 0 100 100" version="1.1" xmlns="http://www.w3.org/2000/svg">`,
            `<circle stroke="#003dd9" fill="${me.fillColor}" stroke-width="10" cx="50" cy="50" r="35"/>`,
            `</svg>`
        ].join('\n');

        me._vertexMarker = new google.maps.Marker({
            position: me.to,
            zIndex: 1000002,
            draggable: true,
            icon: { url: 'data:image/svg+xml;charset=UTF-8,' + encodeURIComponent(svg), scaledSize: new google.maps.Size(15, 15) },
            raiseOnDrag: false,
            optimized: true,

        });

        me._vertexMarker.setMap(me.map);
        me._vertexMarker.setCursor('col-resize');
        me._centerMarker && me._vertexMarkerAddEventListener();
    },

    DrawingManager.prototype._vertexMarkerAddEventListener = function () {
        const me = this;
        me.isMaxDistance = false;
        google.maps.event.addListener(me._vertexMarker, 'drag', (event) => {
            if (!me._centerMarker) { return; }
            const distance = me.getDistanceTo(me._centerMarker.getPosition(), event.latLng);
            if (distance >= me.circleOptions.maxRadius) {

                me._vertexMarker.setOpacity(0);
                me._vertexMarker.setZIndex(0);

                if (me._vertexMarkerCopy) {
                    me._vertexMarkerCopy.setOpacity(1);
                    me._vertexMarkerCopy.setZIndex(1000);
                }

                me.isMaxDistance = true;
                me.map.setOptions({
                    draggable: false,
                });
                google.maps.event.trigger(me, 'draw:circle_radius_max', null);

                return false;
            };

            me.map.setOptions({
                draggable: true,
            });

            me.isMaxDistance = false;
            if (me._vertexMarkerCopy) {
                me._vertexMarkerCopy.setMap(null);
            }
            if (!me.isMaxDistance) {
                me._createVertexMarkerCopy();
            }

            me.radius = distance;

            if (me._circle) {
                me._circle.setRadius(distance);
            }

            me._vertexMarker.setOpacity(1)
            me._vertexMarker.setZIndex(1000);

            if (me._vertexMarkerCopy) {
                me._vertexMarkerCopy.setOpacity(0)
                me._vertexMarkerCopy.setZIndex(0);
            }

            const pixel = me.getXYbyEvent(event);
            const ev = {
                pixel,
                radius: Math.floor(me._circle.getRadius()),
            }
            google.maps.event.trigger(me, 'draw:circle_radius_change', ev);

        });

        const _vertMarker = google.maps.event.addListener(me._vertexMarker, 'mouseover', () => {
            if (me._vertexMarkerCopy) {
                me._vertexMarkerCopy.setOpacity(0)
                me._vertexMarkerCopy.setZIndex(0);
            }

            me._vertexMarker.setOpacity(1);
            me._vertexMarker.setZIndex(1000);
            google.maps.event.removeListener(_vertMarker);
        });

        google.maps.event.addListener(me._vertexMarker, 'dragend', () => {
            if (me._vertexMarkerCopy) {
                if (me.isMaxDistance) {

                    me._vertexMarker.setPosition(me._vertexMarkerCopy.getPosition());
                    me.map.setOptions({
                        draggable: true,
                    });
                    if (me._vertexMarkerCopy) {
                        me._vertexMarkerCopy.setOpacity(0)
                        me._vertexMarkerCopy.setZIndex(0);
                    }

                    me._vertexMarker.setOpacity(1)
                    me._vertexMarker.setZIndex(1000);
                }
            }

            google.maps.event.trigger(me, 'draw:circle_radius_complete', this._getInfo());
        });

    },

    DrawingManager.prototype._createVertexMarkerCopy = function () {

        const me = this;
        me.fillColor = 'white';

        let svg = [
            `<?xml version="1.0"?>`,
            `<svg width="16px" height="16px" viewBox="0 0 100 100" version="1.1" xmlns="http://www.w3.org/2000/svg">`,
            `<circle stroke="#003dd9" fill="${me.fillColor}" stroke-width="10" cx="50" cy="50" r="35"/>`,
            `</svg>`
        ].join('\n');

        me._vertexMarkerCopy = new google.maps.Marker({
            position: me._vertexMarker.getPosition(),
            opacity: 0,
            zIndex: 99,
            draggable: false,
            icon: { url: 'data:image/svg+xml;charset=UTF-8,' + encodeURIComponent(svg), scaledSize: new google.maps.Size(15, 15) },
            raiseOnDrag: false,
            optimized: true,

        });

        me._vertexMarkerCopy.setMap(me.map);
        me._vertexMarkerCopyAddEventListener();

    },

    DrawingManager.prototype._vertexMarkerCopyAddEventListener = function () {
        const me = this;
        const _vertMarkerCopy = me._vertexMarkerCopy.addListener('mouseover', () => {
            if (me._vertexMarkerCopy) {
                me._vertexMarkerCopy.setOpacity(0);
                me._vertexMarkerCopy.setZIndex(0);
            }

            me._vertexMarker.setOpacity(1);
            me._vertexMarker.setZIndex(1000);
            me._vertexMarker.setPosition(me._vertexMarkerCopy.getPosition());
            google.maps.event.removeListener(_vertMarkerCopy);
        });
    },

    DrawingManager.prototype.getXYbyEvent = function (event) {
        const me = this;
        const client = me.getProjection().fromLatLngToContainerPixel(event.latLng);
        return {
            clientX: client.x,
            clientY: client.y
        }
    },

    DrawingManager.prototype._centerMarkerAddEventListener = function () {
        let me = this;
        google.maps.event.addListener(me._centerMarker, 'drag', function (event) {

            me._vertexMarker.setOpacity(0);
            me._vertexMarkerCopy && me._vertexMarkerCopy.setOpacity(0);

            const to = me.destination(event.latLng, 90, me.radius);

            me._vertexMarker && me._vertexMarker.setPosition(to);
            me._vertexMarkerCopy && me._vertexMarkerCopy.setPosition(to);

            google.maps.event.trigger(me, 'draw:circle_centre_change', me._getInfo());
        });

        google.maps.event.addListener(me._centerMarker, 'dragend', function (event) {

            const to = me.destination(event.latLng, 90, me.radius);

            me._vertexMarker && me._vertexMarker.setPosition(to);

            me._vertexMarkerCopy && me._vertexMarkerCopy.setPosition(to);

            me._vertexMarker.setOpacity(1);
            google.maps.event.trigger(me, 'draw:circle_center_complete', me._getInfo());
        });

        google.maps.event.clearListeners(me._centerMarker, 'click');

    },

    DrawingManager.prototype.destination = function (latlng, heading, distance) {
        heading = (heading + 360) % 360;
        var rad = Math.PI / 180,
            radInv = 180 / Math.PI,
            R = 6378137,
            lon1 = latlng.lng() * rad,
            lat1 = latlng.lat() * rad,
            rheading = heading * rad,
            sinLat1 = Math.sin(lat1),
            cosLat1 = Math.cos(lat1),
            cosDistR = Math.cos(distance / R),
            sinDistR = Math.sin(distance / R),
            lat2 = Math.asin(sinLat1 * cosDistR + cosLat1 *
                sinDistR * Math.cos(rheading)),
            lon2 = lon1 + Math.atan2(Math.sin(rheading) * sinDistR *
                cosLat1, cosDistR - sinLat1 * Math.sin(lat2));
        lon2 = lon2 * radInv;
        lon2 = lon2 > 180 ? lon2 - 360 : lon2 < -180 ? lon2 + 360 : lon2;
        return new google.maps.LatLng(lat2 * radInv, lon2)
    },

    DrawingManager.prototype.degreeToRad = function (degree) {
        return Math.PI * degree / 180;
    },

    DrawingManager.prototype._getRange = function (v, a, b) {
        if (a) {
            v = Math.max(v, a);
        }
        if (b) {
            v = Math.min(v, b);
        }
        return v;
    },

    DrawingManager.prototype._getLoop = function (v, a, b) {
        while (v > b) {
            v -= b - a
        }
        while (v < a) {
            v += b - a
        }
        return v;
    },

    DrawingManager.prototype.getDistanceTo = function (point1, point2) {

        const me = this;
        point1.ln = me._getLoop(point1.lng(), -180, 180);
        point1.lt = me._getRange(point1.lat(), -74, 74);
        point2.ln = me._getLoop(point2.lng(), -180, 180);
        point2.lt = me._getRange(point2.lat(), -74, 74);

        let x1, x2, y1, y2;
        x1 = this.degreeToRad(point1.ln);
        y1 = this.degreeToRad(point1.lt);
        x2 = this.degreeToRad(point2.ln);
        y2 = this.degreeToRad(point2.lt);

        return 6370996.81 * Math.acos((Math.sin(y1) * Math.sin(y2) + Math.cos(y1) * Math.cos(y2) * Math.cos(x2 - x1)));
    },

    DrawingManager.prototype.extend = function (obj1, obj2) {
        return (function (object) {
            var property;
            for (property in object.prototype) {
                this.prototype[property] = object.prototype[property];
            }
            return this;
        }).apply(obj1, [obj2]);
    },

    DrawingManager.prototype._getInfo = function () {
        const me = this;
        const position = {
            latitude: me._centerMarker && me._centerMarker.getPosition().lat() || me._circle.getCenter.lat(),
            longitude: me._centerMarker && me._centerMarker.getPosition().lng() || me._circle.getCenter.lng()
        }
        const info = {
            isDraw: me.isDraw,
            bounds: me._circle.getBounds(),
            radius: Math.floor(me._circle.getRadius()),
            position,
            zoom: me.map.getZoom(),
        };

        return info;
    },

    DrawingManager.prototype._getPosition = function () {
        const me = this;
        let position = {
            latitude: me._centerMarker.getPosition().lat(),
            longitude: me._centerMarker.getPosition().lng(),
        }
        let drawMarker = {
            isDraw: me.isDraw,
            position,
            placeId: me._centerMarker.placeId || null,
        }
        return drawMarker;
    };

if (typeof module == 'object') {
    module.exports = { default: DrawingManager, DrawingManager: DrawingManager };
}